/*
 For more information, please see: http://software.sci.utah.edu
 
 The MIT License
 
 Copyright (c) 2014 Scientific Computing and Imaging Institute,
 University of Utah.
 
 
 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
 to deal in the Software without restriction, including without limitation
 the rights to use, copy, modify, merge, publish, distribute, sublicense,
 and/or sell copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.
 */

#include <gtest/gtest.h>

#include <boost/filesystem.hpp>

#include <Core/Datatypes/Field.h> 
#include <Core/Datatypes/Mesh.h> 
#include <Core/Datatypes/FieldInformation.h>

#include <Testing/Util/TestUtilities.h>
#include <Testing/Util/FieldReader.h>

#include <sci_defs/cmake_test_environment_testdefs.h>

using namespace SCIRun;
using namespace SCIRun::TestUtils;

class TriSurfMeshTest : public ::testing::Test
{
protected:
  virtual void SetUp()
  {
    // TODO: triangle (single mesh unit) example

    // cube example
    {
      FieldInformation fi("TriSurfMesh", 1, "double");
      cubeField_ = CreateField(fi);
      VMesh *vmesh = cubeField_->vmesh();  
      VMesh::Node::array_type vdata;
      vdata.resize(3);
      
      vmesh->node_reserve(8);
      vmesh->elem_reserve(12);
      vmesh->add_point( Point(0.0, 1.0, 0.0) );
      vmesh->add_point( Point(0.0, 0.0, 0.0) );
      vmesh->add_point( Point(1.0, 1.0, 0.0) );
      vmesh->add_point( Point(1.0, 0.0, 0.0) );
      vmesh->add_point( Point(1.0, 0.0, -1.0) );
      vmesh->add_point( Point(1.0, 1.0, -1.0) );
      vmesh->add_point( Point(0.0, 1.0, -1.0) );
      vmesh->add_point( Point(0.0, 0.0, -1.0) );
      vdata[0] = 0; vdata[1] = 1; vdata[2] = 7;
      vmesh->add_elem(vdata);
      vdata[0] = 0; vdata[1] = 7; vdata[2] = 6;
      vmesh->add_elem(vdata);
      vdata[0] = 1; vdata[1] = 0; vdata[2] = 2;
      vmesh->add_elem(vdata);
      vdata[0] = 1; vdata[1] = 3; vdata[2] = 2;
      vmesh->add_elem(vdata);
      vdata[0] = 2; vdata[1] = 3; vdata[2] = 4;
      vmesh->add_elem(vdata);
      vdata[0] = 2; vdata[1] = 4; vdata[2] = 5;
      vmesh->add_elem(vdata);
      vdata[0] = 4; vdata[1] = 7; vdata[2] = 1;
      vmesh->add_elem(vdata);
      vdata[0] = 4; vdata[1] = 3; vdata[2] = 1;
      vmesh->add_elem(vdata);
      vdata[0] = 5; vdata[1] = 6; vdata[2] = 0;
      vmesh->add_elem(vdata);
      vdata[0] = 5; vdata[1] = 2; vdata[2] = 0;
      vmesh->add_elem(vdata);
      vdata[0] = 7; vdata[1] = 6; vdata[2] = 5;
      vmesh->add_elem(vdata);
      vdata[0] = 7; vdata[1] = 4; vdata[2] = 5;
      vmesh->add_elem(vdata);
    }

    // "consistent" mesh example from file
    {
      consistentTriSurfMeshField_ = open_scirun_field("consistent-mesh_trisurf.fld");
    }
  }

  FieldHandle cubeField_;
  FieldHandle consistentTriSurfMeshField_;
};

TEST_F(TriSurfMeshTest, CubeSynchronizeNodeNeighbors)
{
  ASSERT_TRUE(cubeField_.get_rep() != 0);
  VMesh *vmesh = cubeField_->vmesh();
  ASSERT_TRUE(vmesh->synchronize(Mesh::NODE_NEIGHBORS_E));
}

TEST_F(TriSurfMeshTest, CubeSynchronizeEdges)
{
  ASSERT_TRUE(cubeField_.get_rep() != 0);
  VMesh *vmesh = cubeField_->vmesh();  
  ASSERT_TRUE(vmesh->synchronize(Mesh::EDGES_E));
}

TEST_F(TriSurfMeshTest, CubeSynchronizeNeighborsEdges)
{
  ASSERT_TRUE(cubeField_.get_rep() != 0);
  VMesh *vmesh = cubeField_->vmesh();
  ASSERT_TRUE(vmesh->synchronize(Mesh::NODE_NEIGHBORS_E|Mesh::EDGES_E));
}

TEST_F(TriSurfMeshTest, CubeGetEdgesFromNode)
{
  ASSERT_TRUE(cubeField_.get_rep() != 0);  
  VMesh *vmesh = cubeField_->vmesh();
  ASSERT_TRUE(vmesh->synchronize(Mesh::NODE_NEIGHBORS_E|Mesh::EDGES_E));

  VMesh::Node::iterator meshNodeIter;
  VMesh::Node::iterator meshNodeEnd;

  
  vmesh->end(meshNodeEnd);
  
  for (vmesh->begin(meshNodeIter); meshNodeIter != meshNodeEnd; ++meshNodeIter)
  {
    VMesh::Node::index_type nodeID = *meshNodeIter;
    VMesh::Edge::array_type edgesFromNode(1);
    Point p;
    
    vmesh->get_edges(edgesFromNode, nodeID);
    vmesh->get_point(p, nodeID);
    
    // cube nodes should have at least 3 edges, and no more than 5
    ASSERT_GE(edgesFromNode.size(), 3);
    ASSERT_LE(edgesFromNode.size(), 5);
  }
}

TEST_F(TriSurfMeshTest, ConsistentSynchronizeNodeNeighbors)
{
  ASSERT_TRUE(consistentTriSurfMeshField_.get_rep() != 0);
  VMesh *vmesh = consistentTriSurfMeshField_->vmesh();
  ASSERT_TRUE(vmesh->synchronize(Mesh::NODE_NEIGHBORS_E));
}

TEST_F(TriSurfMeshTest, ConsistentSynchronizeEdges)
{
  ASSERT_TRUE(consistentTriSurfMeshField_.get_rep() != 0);
  VMesh *vmesh = consistentTriSurfMeshField_->vmesh();  
  ASSERT_TRUE(vmesh->synchronize(Mesh::EDGES_E));
}

TEST_F(TriSurfMeshTest, ConsistentSynchronizeNeighborsEdges)
{
  ASSERT_TRUE(consistentTriSurfMeshField_.get_rep() != 0);
  VMesh *vmesh = consistentTriSurfMeshField_->vmesh();
  ASSERT_TRUE(vmesh->synchronize(Mesh::NODE_NEIGHBORS_E|Mesh::EDGES_E));
}

TEST_F(TriSurfMeshTest, ConsistentGetEdgesFromNode)
{
  ASSERT_TRUE(consistentTriSurfMeshField_.get_rep() != 0);  
  VMesh *vmesh = consistentTriSurfMeshField_->vmesh();
  ASSERT_TRUE(vmesh->synchronize(Mesh::NODE_NEIGHBORS_E|Mesh::EDGES_E));
  
  VMesh::Node::iterator meshNodeIter;
  VMesh::Node::iterator meshNodeEnd;

  vmesh->end(meshNodeEnd);
  
  for (vmesh->begin(meshNodeIter); meshNodeIter != meshNodeEnd; ++ meshNodeIter)
  {
    VMesh::Node::index_type nodeID = *meshNodeIter;
    VMesh::Edge::array_type edgesFromNode(1);
    Point p;
    
    vmesh->get_edges(edgesFromNode, nodeID);
    vmesh->get_point(p, nodeID);

    // mesh nodes should have at least 4 edges, and no more than 6
    ASSERT_GE(edgesFromNode.size(), 4);
    ASSERT_LE(edgesFromNode.size(), 6);

    // quick and dirty check...
//    std::cout << "Node " << nodeID << " (point=" << p.get_string() << ") edges=[" << join(edgesFromNode) << "], "
//    << edgesFromNode.size() << " edges at node." << std::endl;
  }
}
