/*
 For more information, please see: http://software.sci.utah.edu
 
 The MIT License
 
 Copyright (c) 2014 Scientific Computing and Imaging Institute,
 University of Utah.
 
 
 Permission is hereby granted, free of charge, to any person obtaining a
 copy of this software and associated documentation files (the "Software"),
 to deal in the Software without restriction, including without limitation
 the rights to use, copy, modify, merge, publish, distribute, sublicense,
 and/or sell copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 DEALINGS IN THE SOFTWARE.
 */

#include <gtest/gtest.h>

#include <Core/Datatypes/Field.h> 
#include <Core/Datatypes/FieldInformation.h>

#include <Testing/Util/TestUtilities.h>

#include <vector>

using namespace SCIRun;
using namespace SCIRun::TestUtils;

class VFieldTest : public ::testing::Test
{
protected:
  virtual void SetUp()
  {
    {
      FieldInformation fi("TriSurfMesh", 0, "double");
      emptyTriSurfFieldConstantBasis_ = CreateField(fi);
    }

    {
      FieldInformation fi("TriSurfMesh", 1, "double");
      emptyTriSurfFieldLinearBasis_ = CreateField(fi);
    }

    // tetrahedron (single mesh unit) examples
    {
      FieldInformation fi("TetVolMesh", 0, "double");
      singleTetFieldConstantBasis_ = CreateField(fi);
      VMesh *vmesh = singleTetFieldConstantBasis_->vmesh();
      VMesh::Node::array_type vdata;
      vdata.resize(4);
      
      vmesh->node_reserve(4);
      vmesh->elem_reserve(1);
      vmesh->add_point( Point(0, 0, 0) );
      vmesh->add_point( Point(0.5, 1.0, 0) );
      vmesh->add_point( Point(1.0, 0, 0) );
      vmesh->add_point( Point(0.5, 0.5, 1.0) );
      for (size_type i = 0; i < 4; ++i)
      {
        vdata[i] = i;
      }
      vmesh->add_elem(vdata);
    }

    {
      FieldInformation fi("TetVolMesh", 1, "double");
      singleTetFieldLinearBasis_ = CreateField(fi);
      VMesh *vmesh = singleTetFieldLinearBasis_->vmesh();
      VMesh::Node::array_type vdata;
      vdata.resize(4);
      
      vmesh->node_reserve(4);
      vmesh->elem_reserve(1);
      vmesh->add_point( Point(0, 0, 0) );
      vmesh->add_point( Point(0.5, 1.0, 0) );
      vmesh->add_point( Point(1.0, 0, 0) );
      vmesh->add_point( Point(0.5, 0.5, 1.0) );
      for (size_type i = 0; i < 4; ++i)
      {
        vdata[i] = i;
      }
      vmesh->add_elem(vdata);
    }
  }
  FieldHandle emptyTriSurfFieldConstantBasis_;
  FieldHandle emptyTriSurfFieldLinearBasis_;
  FieldHandle singleTetFieldConstantBasis_;
  FieldHandle singleTetFieldLinearBasis_;
};

TEST_F(VFieldTest, EmptyFieldConstantBasis)
{
  ASSERT_TRUE(emptyTriSurfFieldConstantBasis_.get_rep() != 0);
  
  VMesh *vmesh = emptyTriSurfFieldConstantBasis_->vmesh();
  VField *vfield = emptyTriSurfFieldConstantBasis_->vfield();
  
  ASSERT_EQ(vfield->num_values(), 0);
  
//  VMesh::dimension_type dim;
//  vfield->get_values_dimension(dim);
//
//std::cerr << dim[0] << ", " << dim[1] << ", " << dim[2] << std::endl;
//  
//  ASSERT_EQ(dim[0], 0);
//  ASSERT_EQ(dim[1], 0);
//  ASSERT_EQ(dim[2], 0);
}

TEST_F(VFieldTest, EmptyFieldAddValuesConstantBasis)
{
  ASSERT_TRUE(emptyTriSurfFieldConstantBasis_.get_rep() != 0);
  
  VMesh *vmesh = emptyTriSurfFieldConstantBasis_->vmesh();
  VField *vfield = emptyTriSurfFieldConstantBasis_->vfield();
  vfield->resize_values();
  
  std::vector<double> values(1);
  values[0] = 10.0;
  vfield->set_values(values);
  
  ASSERT_EQ(vfield->num_values(), 0);
}

TEST_F(VFieldTest, EmptyFieldLinearBasis)
{
  ASSERT_TRUE(emptyTriSurfFieldLinearBasis_.get_rep() != 0);

  VMesh *vmesh = emptyTriSurfFieldLinearBasis_->vmesh();
  VField *vfield = emptyTriSurfFieldLinearBasis_->vfield();
  
  ASSERT_EQ(vfield->num_values(), 0);
}

TEST_F(VFieldTest, EmptyFieldAddValuesLinearBasis)
{
  ASSERT_TRUE(emptyTriSurfFieldLinearBasis_.get_rep() != 0);
  
  VMesh *vmesh = emptyTriSurfFieldLinearBasis_->vmesh();
  VField *vfield = emptyTriSurfFieldLinearBasis_->vfield();
  vfield->resize_values();

  std::vector<double> values(1);
  values[0] = 10.0;
  vfield->set_values(values);

  ASSERT_EQ(vfield->num_values(), 0);
}

TEST_F(VFieldTest, TetVolMeshAddValuesConstantBasis)
{
  ASSERT_TRUE(singleTetFieldConstantBasis_.get_rep() != 0);
  
  VMesh *vmesh = singleTetFieldConstantBasis_->vmesh();
  VField *vfield = singleTetFieldConstantBasis_->vfield();
  vfield->resize_values();

  std::vector<double> values(1);
  values[0] = 1.0;
  vfield->set_values(values);
  
  ASSERT_EQ(vfield->num_values(), 1);
}


TEST_F(VFieldTest, TetVolMeshAddValuesLinearBasis)
{
  ASSERT_TRUE(singleTetFieldLinearBasis_.get_rep() != 0);
  
  VMesh *vmesh = singleTetFieldLinearBasis_->vmesh();
  VField *vfield = singleTetFieldLinearBasis_->vfield();
  vfield->resize_values();

  std::vector<double> values(4);
  values[0] = 1.0;
  values[1] = 2.0;
  values[2] = 3.0;
  values[3] = 4.0;
  vfield->set_values(values);
  
  ASSERT_EQ(vfield->num_values(), 4);
}