/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/


/* 
 * tkCanvBLine.c --
 *
 *	This file implements beveled line items for canvas widgets.
 *
 */

#include <stdio.h>
#include <tkPort.h>
#include <tkInt.h>
#include <tcl.h>

#if defined(_WIN32) && !defined(BUILD_SCIRUN_STATIC)
#define SCISHARE __declspec(dllexport)
#else
#define SCISHARE
#endif

void Tk_DrawBeveledLine(Display* display, Drawable drawable,
			Tk_3DBorder border, XPoint *pointPtr,
			int numPoints, int width, int borderWidth,
			int relief);

/*
 * The structure below defines the record for each line item.
 */

typedef struct BLineItem  {
  Tk_Item header;		/* Generic stuff that's the same for all
       * types.  MUST BE FIRST IN STRUCTURE. */
  int numPoints;		/* Number of points in line (always >= 2). */
  double *coordPtr;		/* Pointer to malloc-ed array containing
       * x- and y-coords of all points in line.
       * X-coords are even-valued indices, y-coords
       * are corresponding odd-valued indices. If
       * the line has arrowheads then the first
       * and last points have been adjusted to refer
       * to the necks of the arrowheads rather than
       * their tips.  The actual endpoints are
       * stored in the *firstArrowPtr and
       * *lastArrowPtr, if they exist. */
  int width;			/* Width of line. */
  int capStyle;		/* Cap style for line. */
  int joinStyle;		/* Join style for line. */
  Tk_3DBorder border;		/* Graphics context for filling line. */
  int borderWidth;
  int relief;
} BLineItem;

/*
 * Prototypes for procedures defined in this file:
 */

static void		ComputeBLineBbox (Tk_Canvas canvas, BLineItem *linePtr);
static int		ConfigureBLine (Tcl_Interp *interp, Tk_Canvas canvas, 
                              Tk_Item *itemPtr, int argc,
                              CONST84 char **argv, int flags);
static int		CreateBLine (Tcl_Interp *interp, Tk_Canvas canvas,
                           struct Tk_Item *itemPtr, int argc, 
                           CONST84 char **argv);
                           
static void		DeleteBLine (Tk_Canvas canvas, Tk_Item *itemPtr, 
                           Display *display);
static void		DisplayBLine (Tk_Canvas canvas,
                            Tk_Item *itemPtr, Display *display, 
                            Drawable dst,
                            int x, int y, int width, int height);
static int		BLineCoords (Tcl_Interp *interp,
                           Tk_Canvas canvas, Tk_Item *itemPtr,
                           int argc,  CONST84 char **argv);

static int		BLineToArea (Tk_Canvas canvas, Tk_Item *itemPtr, double *rectPtr);
static double	BLineToPoint (Tk_Canvas canvas,Tk_Item *itemPtr, double *coordPtr);

static int		BLineToPostscript (Tcl_Interp *interp, Tk_Canvas canvas, 
                                 Tk_Item *itemPtr, int prepass);
static void		ScaleBLine (Tk_Canvas canvas, Tk_Item *itemPtr, 
                          double originX, double originY,
                          double scaleX, double scaleY);
static void		TranslateBLine (Tk_Canvas canvas, Tk_Item *itemPtr, 
                              double deltaX, double deltaY);

/*
 * Information used for parsing configuration specs.  If you change any
 * of the default strings, be sure to change the corresponding default
 * values in CreateBLine.
 */

// initialize in createBLine
static Tk_CustomOption tagsOption;
//= {Tk_CanvasTagsParseProc,
//                                     Tk_CanvasTagsPrintProc, (ClientData) NULL};

static Tk_ConfigSpec configSpecs[] = {
  {TK_CONFIG_CAP_STYLE, "-capstyle", (char *) NULL, (char *) NULL,
      "butt", Tk_Offset(BLineItem, capStyle), TK_CONFIG_DONT_SET_DEFAULT},
  {TK_CONFIG_BORDER, "-fill", (char *) NULL, (char *) NULL,
      "black", Tk_Offset(BLineItem, border), TK_CONFIG_NULL_OK},
  {TK_CONFIG_JOIN_STYLE, "-joinstyle", (char *) NULL, (char *) NULL,
      "round", Tk_Offset(BLineItem, joinStyle), TK_CONFIG_DONT_SET_DEFAULT},
  {TK_CONFIG_CUSTOM, "-tags", (char *) NULL, (char *) NULL,
      (char *) NULL, 0, TK_CONFIG_NULL_OK, &tagsOption},
  {TK_CONFIG_PIXELS, "-width", (char *) NULL, (char *) NULL,
      "1", Tk_Offset(BLineItem, width), TK_CONFIG_DONT_SET_DEFAULT},
  {TK_CONFIG_PIXELS, "-borderwidth", (char *) NULL, (char *) NULL,
      "2", Tk_Offset(BLineItem, borderWidth), TK_CONFIG_DONT_SET_DEFAULT},
  {TK_CONFIG_RELIEF, "-relief", (char*) NULL, (char *) NULL,
      "raised", Tk_Offset(BLineItem, relief), TK_CONFIG_DONT_SET_DEFAULT},
  {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
      (char *) NULL, 0, 0}
};

/*
 * The structures below defines the line item type by means
 * of procedures that can be invoked by generic item code.
 */

Tk_ItemType TkBLineType = {
  "bline",				/* name */
  sizeof(BLineItem),			/* itemSize */
  (Tk_ItemCreateProc *)CreateBLine,   /* createProc */
  configSpecs,			/* configSpecs */
  (Tk_ItemConfigureProc *)ConfigureBLine, /* configureProc */
  (Tk_ItemCoordProc *)BLineCoords,	/* coordProc */
  DeleteBLine,			/* deleteProc */
  DisplayBLine,			/* displayProc */
  0,					/* alwaysRedraw */
  BLineToPoint,			/* pointProc */
  BLineToArea,			/* areaProc */
  BLineToPostscript,			/* postscriptProc */
  ScaleBLine,				/* scaleProc */
  TranslateBLine,			/* translateProc */
  (Tk_ItemIndexProc *) NULL,		/* indexProc */
  (Tk_ItemCursorProc *) NULL,		/* icursorProc */
  (Tk_ItemSelectionProc *) NULL,	/* selectionProc */
  (Tk_ItemInsertProc *) NULL,		/* insertProc */
  (Tk_ItemDCharsProc *) NULL,		/* dTextProc */
  (Tk_ItemType *) NULL		/* nextPtr */
};

/*
 * The definition below determines how large are static arrays
 * used to hold spline points (splines larger than this have to
 * have their arrays malloc-ed).
 */

#define MAX_STATIC_POINTS 200

/*
 *--------------------------------------------------------------
 *
 * CreateBLine --
 *
 *	This procedure is invoked to create a new line item in
 *	a canvas.
 *
 * Results:
 *	A standard Tcl return value.  If an error occurred in
 *	creating the item, then an error message is left in
 *	canvasPtr->interp->result;  in this case itemPtr is
 *	left uninitialized, so it can be safely freed by the
 *	caller.
 *
 * Side effects:
 *	A new line item is created.
 *
 *--------------------------------------------------------------
 */

static int
CreateBLine(Tcl_Interp* interp, Tk_Canvas canvas, Tk_Item *itemPtr, 
            int argc, CONST84 char** argv)
{
  BLineItem *linePtr = (BLineItem *) itemPtr;
  int i;

  tagsOption.parseProc = Tk_CanvasTagsParseProc;
  tagsOption.printProc = Tk_CanvasTagsPrintProc;

  if (argc < 4) 
  {
    Tcl_AppendResult(interp, "wrong # args:  should be \"",
      Tk_PathName(Tk_CanvasTkwin(canvas)), "\" create ",
      itemPtr->typePtr->name, " x1 y1 x2 y2 ?x3 y3 ...? ?options?",
      (char *) NULL);
    return TCL_ERROR;
  }

  /*
   * Carry out initialization that is needed to set defaults and to
   * allow proper cleanup after errors during the the remainder of
   * this procedure.
   */

  linePtr->numPoints = 0;
  linePtr->coordPtr = NULL;
  linePtr->width = 1;
  linePtr->capStyle = CapButt;
  linePtr->joinStyle = JoinRound;
  linePtr->borderWidth = 2;
  linePtr->relief = TK_RELIEF_RAISED;
  linePtr->border = NULL;

  /*
   * Count the number of points and then parse them into a point
   * array.  Leading arguments are assumed to be points if they
   * start with a digit or a minus sign followed by a digit.
   */

  for (i = 4; i < (argc-1); i+=2) 
  {
    if ((!isdigit(UCHAR(argv[i][0]))) &&
        ((argv[i][0] != '-') || (!isdigit(UCHAR(argv[i][1]))))) 
    {
      break;
    }
  }
  if (BLineCoords(interp, canvas, itemPtr, i, argv) != TCL_OK) 
  {
    goto error;
  }
  if (ConfigureBLine(interp, canvas, itemPtr, argc-i, argv+i, 0) == TCL_OK) 
  {
    return TCL_OK;
  }

  error:
  DeleteBLine(canvas, itemPtr, Tk_Display(Tk_CanvasTkwin(canvas)));
  return TCL_ERROR;
}

/*
 *--------------------------------------------------------------
 *
 * BLineCoords --
 *
 *	This procedure is invoked to process the "coords" widget
 *	command on lines.  See the user documentation for details
 *	on what it does.
 *
 * Results:
 *	Returns TCL_OK or TCL_ERROR, and sets canvasPtr->interp->result.
 *
 * Side effects:
 *	The coordinates for the given item may be changed.
 *
 *--------------------------------------------------------------
 */

static int
BLineCoords(Tcl_Interp *interp, Tk_Canvas canvas, Tk_Item *itemPtr, 
            int argc, CONST84 char** argv)
{
  register BLineItem *linePtr = (BLineItem *) itemPtr;
  char buffer[TCL_DOUBLE_SPACE];
  int i, numPoints;

  if (argc == 0) 
  {
    double *coordPtr;
    int numCoords;

    numCoords = 2*linePtr->numPoints;
    coordPtr = linePtr->coordPtr;
    for (i = 0; i < numCoords; i++, coordPtr++) 
    {
        if (i == 2) 
        {
          coordPtr = linePtr->coordPtr+2;
        }
        Tcl_PrintDouble(interp, *coordPtr, buffer);
        Tcl_AppendElement(interp, buffer);
      }
    } 
    else if (argc < 4) 
    {
      Tcl_AppendResult(interp,
        "too few coordinates for line:  must have at least 4",
        (char *) NULL);
      return TCL_ERROR;
    } 
    else if (argc & 1) 
    {
      Tcl_AppendResult(interp,
        "odd number of coordinates specified for line",
        (char *) NULL);
      return TCL_ERROR;
    } 
    else 
    {
      numPoints = argc/2;
      if (linePtr->numPoints != numPoints) 
      {
        if (linePtr->coordPtr != NULL) 
        {
          ckfree((char *) linePtr->coordPtr);
        }
        linePtr->coordPtr = (double *) ckalloc((unsigned)(sizeof(double) * argc));
        linePtr->numPoints = numPoints;
      }
      for (i = argc-1; i >= 0; i--) 
      {
        if (Tk_CanvasGetCoord(interp, canvas, argv[i],
            &linePtr->coordPtr[i]) != TCL_OK) 
        {
          return TCL_ERROR;
        }
      }

      ComputeBLineBbox(canvas, linePtr);
    }
    return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * ConfigureBLine --
 *
 *	This procedure is invoked to configure various aspects
 *	of a line item such as its background color.
 *
 * Results:
 *	A standard Tcl result code.  If an error occurs, then
 *	an error message is left in canvasPtr->interp->result.
 *
 * Side effects:
 *	Configuration information, such as colors and stipple
 *	patterns, may be set for itemPtr.
 *
 *--------------------------------------------------------------
 */

static int
ConfigureBLine(Tcl_Interp *interp, Tk_Canvas canvas, Tk_Item *itemPtr, 
              int argc, CONST84 char** argv, int flags)
{
  register BLineItem *linePtr = (BLineItem *) itemPtr;
  Tk_Window tkwin;
  
  tkwin = Tk_CanvasTkwin(canvas);
  if (Tk_ConfigureWidget(interp, tkwin, configSpecs, argc, argv,
     (char *) linePtr, flags) != TCL_OK) 
  {
    return TCL_ERROR;
  }

  /*
   * Recompute bounding box for line.
   */

  ComputeBLineBbox(canvas, linePtr);

  return TCL_OK;
}

/*
 *--------------------------------------------------------------
 *
 * DeleteBLine --
 *
 *	This procedure is called to clean up the data structure
 *	associated with a line item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Resources associated with itemPtr are released.
 *
 *--------------------------------------------------------------
 */

static void
DeleteBLine(Tk_Canvas canvas, Tk_Item *itemPtr, Display *display)
{
  BLineItem *linePtr = (BLineItem *) itemPtr;

  if (linePtr->coordPtr != NULL) 
  {
    ckfree((char *) linePtr->coordPtr);
  }
  if (linePtr->border != NULL) 
  {
    Tk_Free3DBorder(linePtr->border);
  }
}

/*
 *--------------------------------------------------------------
 *
 * ComputeBLineBbox --
 *
 *	This procedure is invoked to compute the bounding box of
 *	all the pixels that may be drawn as part of a line.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The fields x1, y1, x2, and y2 are updated in the header
 *	for itemPtr.
 *
 *--------------------------------------------------------------
 */

static void
ComputeBLineBbox(Tk_Canvas canvas, BLineItem *linePtr)
{
  register double *coordPtr;
  int i;

  coordPtr = linePtr->coordPtr;
  linePtr->header.x1 = linePtr->header.x2 = (int)coordPtr[0];
  linePtr->header.y1 = linePtr->header.y2 = (int)coordPtr[1];

  /*
   * Compute the bounding box of all the points in the line,
   * then expand in all directions by the line's width to take
   * care of butting or rounded corners and projecting or
   * rounded caps.  This expansion is an overestimate (worst-case
   * is square root of two over two) but it's simple.  Don't do
   * anything special for curves.  This causes an additional
   * overestimate in the bounding box, but is faster.
   */

  for (i = 1, coordPtr = linePtr->coordPtr+2; i < linePtr->numPoints;
    i++, coordPtr += 2) 
  {
    TkIncludePoint((Tk_Item *) linePtr, coordPtr);
  }
  linePtr->header.x1 -= linePtr->width;
  linePtr->header.x2 += linePtr->width;
  linePtr->header.y1 -= linePtr->width;
  linePtr->header.y2 += linePtr->width;

  /*
   * For mitered lines, make a second pass through all the points.
   * Compute the locations of the two miter vertex points and add
   * those into the bounding box.
   */

  if (linePtr->joinStyle == JoinMiter) 
  {
    for (i = linePtr->numPoints, coordPtr = linePtr->coordPtr; i >= 3;
      i--, coordPtr += 2) 
    {
      double miter[4];
      int j;
    
      if (TkGetMiterPoints(coordPtr, coordPtr+2, coordPtr+4,
        (double) linePtr->width, miter, miter+2)) 
      {
        for (j = 0; j < 4; j += 2) 
        {
          TkIncludePoint((Tk_Item *) linePtr, miter+j);
        }
      }
    }
  }

  /*
   * Add one more pixel of fudge factor just to be safe (e.g.
   * X may round differently than we do).
   */

  linePtr->header.x1 -= 3;
  linePtr->header.x2 += 3;
  linePtr->header.y1 -= 3;
  linePtr->header.y2 += 3;
}

/*
 *--------------------------------------------------------------
 *
 * DisplayBLine --
 *
 *	This procedure is invoked to draw a line item in a given
 *	drawable.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	ItemPtr is drawn in drawable using the transformation
 *	information in canvasPtr.
 *
 *--------------------------------------------------------------
 */

static void
DisplayBLine(Tk_Canvas canvas, Tk_Item *itemPtr, Display *display, 
            Drawable drawable, int x, int y, int width, int height)
{
  register BLineItem *linePtr = (BLineItem *) itemPtr;
  XPoint staticPoints[MAX_STATIC_POINTS];
  XPoint *pointPtr;
  register XPoint *pPtr;
  register double *coordPtr;
  int i, numPoints;

  /*
   * Build up an array of points in screen coordinates.  Use a
   * static array unless the line has an enormous number of points;
   * in this case, dynamically allocate an array.  For smoothed lines,
   * generate the curve points on each redisplay.
   */

  numPoints = linePtr->numPoints;

  if (numPoints <= MAX_STATIC_POINTS) 
  {
    pointPtr = staticPoints;
  } 
  else 
  {
    pointPtr = (XPoint *) ckalloc((unsigned) (numPoints * sizeof(XPoint)));
  }

  for (i = 0, coordPtr = linePtr->coordPtr, pPtr = pointPtr;
    i < linePtr->numPoints;  i += 1, coordPtr += 2, pPtr++) 
  {
    Tk_CanvasDrawableCoords(canvas, coordPtr[0], coordPtr[1],
      &pPtr->x, &pPtr->y);
  }

  Tk_DrawBeveledLine(display, drawable, linePtr->border,
		       pointPtr, numPoints, linePtr->width,
		       linePtr->borderWidth, linePtr->relief);
    
  if (pointPtr != staticPoints) 
  {
    ckfree((char *) pointPtr);
  }
}

/*
 *--------------------------------------------------------------
 *
 * BLineToPoint --
 *
 *	Computes the distance from a given point to a given
 *	line, in canvas units.
 *
 * Results:
 *	The return value is 0 if the point whose x and y coordinates
 *	are pointPtr[0] and pointPtr[1] is inside the line.  If the
 *	point isn't inside the line then the return value is the
 *	distance from the point to the line.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

	/* ARGSUSED */
static double
BLineToPoint(Tk_Canvas canvas, Tk_Item *itemPtr, double *pointPtr)
{
  register BLineItem *linePtr = (BLineItem *) itemPtr;
  register double *coordPtr, *linePoints;
  double staticSpace[2*MAX_STATIC_POINTS];
  double poly[10];
  double bestDist, dist;
  int numPoints, count;
  int changedMiterToBevel;	/* Non-zero means that a mitered corner
       * had to be treated as beveled after all
       * because the angle was < 11 degrees. */

  bestDist = 1.0e40;

  /*
   * Handle smoothed lines by generating an expanded set of points
   * against which to do the check.
   */

  numPoints = linePtr->numPoints;
  linePoints = linePtr->coordPtr;

  /*
   * The overall idea is to iterate through all of the edges of
   * the line, computing a polygon for each edge and testing the
   * point against that polygon.  In addition, there are additional
   * tests to deal with rounded joints and caps.
   */

  changedMiterToBevel = 0;
  for (count = numPoints, coordPtr = linePoints; count >= 2;
    count--, coordPtr += 2) 
  {

    /*
     * If rounding is done around the first point then compute
     * the distance between the point and the point.
     */

    if (((linePtr->capStyle == CapRound) && (count == numPoints))
      || ((linePtr->joinStyle == JoinRound)
        && (count != numPoints))) 
    {
      dist = hypot(coordPtr[0] - pointPtr[0], coordPtr[1] - pointPtr[1])
            - linePtr->width/2.0;
      if (dist <= 0.0) 
      {
        bestDist = 0.0;
        goto done;
      } 
      else if (dist < bestDist) 
      {
        bestDist = dist;
      }
    }

    /*
     * Compute the polygonal shape corresponding to this edge,
     * consisting of two points for the first point of the edge
     * and two points for the last point of the edge.
     */

    if (count == numPoints) 
    {
      TkGetButtPoints(coordPtr+2, coordPtr, (double) linePtr->width,
          linePtr->capStyle == CapProjecting, poly, poly+2);
    } 
    else if ((linePtr->joinStyle == JoinMiter) && !changedMiterToBevel) 
    {
      poly[0] = poly[6];
      poly[1] = poly[7];
      poly[2] = poly[4];
      poly[3] = poly[5];
    } 
    else 
    {
      TkGetButtPoints(coordPtr+2, coordPtr, (double) linePtr->width, 0,
        poly, poly+2);

	    /*
	     * If this line uses beveled joints, then check the distance
	     * to a polygon comprising the last two points of the previous
	     * polygon and the first two from this polygon;  this checks
	     * the wedges that fill the mitered joint.
	     */

	    if ((linePtr->joinStyle == JoinBevel) || changedMiterToBevel) 
      {
        poly[8] = poly[0];
        poly[9] = poly[1];
        dist = TkPolygonToPoint(poly, 5, pointPtr);
        if (dist <= 0.0) 
        {
          bestDist = 0.0;
          goto done;
        } 
        else if (dist < bestDist) 
        {
          bestDist = dist;
        }
        changedMiterToBevel = 0;
	    }
    }
    if (count == 2) 
    {
	    TkGetButtPoints(coordPtr, coordPtr+2, (double) linePtr->width,
		    linePtr->capStyle == CapProjecting, poly+4, poly+6);
    } 
    else if (linePtr->joinStyle == JoinMiter) 
    {
	    if (TkGetMiterPoints(coordPtr, coordPtr+2, coordPtr+4,
		    (double) linePtr->width, poly+4, poly+6) == 0) 
      {
        changedMiterToBevel = 1;
        TkGetButtPoints(coordPtr, coordPtr+2, (double) linePtr->width,
          0, poly+4, poly+6);
	    }
    } 
    else 
    {
	    TkGetButtPoints(coordPtr, coordPtr+2, (double) linePtr->width, 0,
		    poly+4, poly+6);
    }
    poly[8] = poly[0];
    poly[9] = poly[1];
    dist = TkPolygonToPoint(poly, 5, pointPtr);
    if (dist <= 0.0) 
    {
      bestDist = 0.0;
      goto done;
    } 
    else if (dist < bestDist) 
    {
      bestDist = dist;
    }
  }

  /*
   * If caps are rounded, check the distance to the cap around the
   * final end point of the line.
   */

  if (linePtr->capStyle == CapRound) 
  {
    dist = hypot(coordPtr[0] - pointPtr[0], coordPtr[1] - pointPtr[1])
      - linePtr->width/2.0;
    if (dist <= 0.0) 
    {
	    bestDist = 0.0;
	    goto done;
    } 
    else if (dist < bestDist) 
    {
	    bestDist = dist;
    }
  }

  done:
  if ((linePoints != staticSpace) && (linePoints != linePtr->coordPtr)) 
  {
    ckfree((char *) linePoints);
  }
  return bestDist;
}

/*
 *--------------------------------------------------------------
 *
 * BLineToArea --
 *
 *	This procedure is called to determine whether an item
 *	lies entirely inside, entirely outside, or overlapping
 *	a given rectangular area.
 *
 * Results:
 *	-1 is returned if the item is entirely outside the
 *	area, 0 if it overlaps, and 1 if it is entirely
 *	inside the given area.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

	/* ARGSUSED */
static int
BLineToArea(Tk_Canvas canvas, Tk_Item *itemPtr, double *rectPtr)
{
  register BLineItem *linePtr = (BLineItem *) itemPtr;
  register double *coordPtr;
  double staticSpace[2*MAX_STATIC_POINTS];
  double *linePoints, poly[10];
  double radius;
  int numPoints, count;
  int changedMiterToBevel;	/* Non-zero means that a mitered corner
       * had to be treated as beveled after all
       * because the angle was < 11 degrees. */
  int inside;			/* Tentative guess about what to return,
       * based on all points seen so far:  one
       * means everything seen so far was
       * inside the area;  -1 means everything
       * was outside the area.  0 means overlap
       * has been found. */ 

  radius = linePtr->width/2.0;
  inside = -1;

  numPoints = linePtr->numPoints;
  linePoints = linePtr->coordPtr;

  coordPtr = linePoints;
  if ((coordPtr[0] >= rectPtr[0]) && (coordPtr[0] <= rectPtr[2])
    && (coordPtr[1] >= rectPtr[1]) && (coordPtr[1] <= rectPtr[3])) 
  {
    inside = 1;
  }

  /*
   * Iterate through all of the edges of the line, computing a polygon
   * for each edge and testing the area against that polygon.  In
   * addition, there are additional tests to deal with rounded joints
   * and caps.
   */

  changedMiterToBevel = 0;
  for (count = numPoints; count >= 2; count--, coordPtr += 2) 
  {

    /*
     * If rounding is done around the first point of the edge
     * then test a circular region around the point with the
     * area.
     */

    if (((linePtr->capStyle == CapRound) && (count == numPoints))
      || ((linePtr->joinStyle == JoinRound)
      && (count != numPoints))) 
    {
      poly[0] = coordPtr[0] - radius;
      poly[1] = coordPtr[1] - radius;
      poly[2] = coordPtr[0] + radius;
      poly[3] = coordPtr[1] + radius;
      if (TkOvalToArea(poly, rectPtr) != inside) 
      {
        inside = 0;
        goto done;
	    }
    }

    /*
     * Compute the polygonal shape corresponding to this edge,
     * consisting of two points for the first point of the edge
     * and two points for the last point of the edge.
     */

    if (count == numPoints) 
    {
      TkGetButtPoints(coordPtr+2, coordPtr, (double) linePtr->width,
          linePtr->capStyle == CapProjecting, poly, poly+2);
    } 
    else if ((linePtr->joinStyle == JoinMiter) && !changedMiterToBevel) 
    {
      poly[0] = poly[6];
      poly[1] = poly[7];
      poly[2] = poly[4];
      poly[3] = poly[5];
    } 
    else 
    {
      TkGetButtPoints(coordPtr+2, coordPtr, (double) linePtr->width, 0,
          poly, poly+2);

	    /*
	     * If the last joint was beveled, then also check a
	     * polygon comprising the last two points of the previous
	     * polygon and the first two from this polygon;  this checks
	     * the wedges that fill the beveled joint.
	     */

	    if ((linePtr->joinStyle == JoinBevel) || changedMiterToBevel) 
      {
        poly[8] = poly[0];
        poly[9] = poly[1];
        if (TkPolygonToArea(poly, 5, rectPtr) != inside) 
        {
          inside = 0;
          goto done;
        }
        changedMiterToBevel = 0;
	    }
    }
    if (count == 2) 
    {
	    TkGetButtPoints(coordPtr, coordPtr+2, (double) linePtr->width,
		    linePtr->capStyle == CapProjecting, poly+4, poly+6);
    }
    else if (linePtr->joinStyle == JoinMiter) 
    {
	    if (TkGetMiterPoints(coordPtr, coordPtr+2, coordPtr+4,
		    (double) linePtr->width, poly+4, poly+6) == 0) 
      {
        changedMiterToBevel = 1;
        TkGetButtPoints(coordPtr, coordPtr+2, (double) linePtr->width,
          0, poly+4, poly+6);
	    }
    } 
    else 
    {
	    TkGetButtPoints(coordPtr, coordPtr+2, (double) linePtr->width, 0,
		    poly+4, poly+6);
    }
    poly[8] = poly[0];
    poly[9] = poly[1];
    if (TkPolygonToArea(poly, 5, rectPtr) != inside) 
    {
      inside = 0;
      goto done;
    }
  }

  /*
   * If caps are rounded, check the cap around the final point
   * of the line.
   */

  if (linePtr->capStyle == CapRound) 
  {
    poly[0] = coordPtr[0] - radius;
    poly[1] = coordPtr[1] - radius;
    poly[2] = coordPtr[0] + radius;
    poly[3] = coordPtr[1] + radius;
    if (TkOvalToArea(poly, rectPtr) != inside) 
    {
	    inside = 0;
	    goto done;
    }
  }

done:
  if ((linePoints != staticSpace) && (linePoints != linePtr->coordPtr)) 
  {
    ckfree((char *) linePoints);
  }
  return inside;
}

/*
 *--------------------------------------------------------------
 *
 * ScaleBLine --
 *
 *	This procedure is invoked to rescale a line item.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The BLine referred to by itemPtr is rescaled so that the
 *	following transformation is applied to all point
 *	coordinates:
 *		x' = originX + scaleX*(x-originX)
 *		y' = originY + scaleY*(y-originY)
 *
 *--------------------------------------------------------------
 */

static void
ScaleBLine(Tk_Canvas canvas, Tk_Item *itemPtr, double originX, double originY, 
           double scaleX, double scaleY)
{
  BLineItem *linePtr = (BLineItem *) itemPtr;
  register double *coordPtr;
  int i;

  for (i = 0, coordPtr = linePtr->coordPtr; i < linePtr->numPoints;
    i++, coordPtr += 2) 
  {
    coordPtr[0] = originX + scaleX*(*coordPtr - originX);
    coordPtr[1] = originY + scaleY*(coordPtr[1] - originY);
  }
  ComputeBLineBbox(canvas, linePtr);
}

/*
 *--------------------------------------------------------------
 *
 * TranslateBLine --
 *
 *	This procedure is called to move a line by a given amount.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The position of the line is offset by (xDelta, yDelta), and
 *	the bounding box is updated in the generic part of the item
 *	structure.
 *
 *--------------------------------------------------------------
 */

static void
TranslateBLine(Tk_Canvas canvas, Tk_Item *itemPtr, double deltaX, double deltaY)
{
  BLineItem *linePtr = (BLineItem *) itemPtr;
  register double *coordPtr;
  int i;

  for (i = 0, coordPtr = linePtr->coordPtr; i < linePtr->numPoints;
    i++, coordPtr += 2) 
  {
    coordPtr[0] += deltaX;
    coordPtr[1] += deltaY;
  }
  ComputeBLineBbox(canvas, linePtr);
}

/*
 *--------------------------------------------------------------
 *
 * LineToPostscript --
 *
 *	This procedure is called to generate Postscript for
 *	line items.
 *
 * Results:
 *	The return value is a standard Tcl result.  If an error
 *	occurs in generating Postscript then an error message is
 *	left in canvasPtr->interp->result, replacing whatever used
 *	to be there.  If no error occurs, then Postscript for the
 *	item is appended to the result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
BLineToPostscript(Tcl_Interp *interp, Tk_Canvas canvas, 
                  Tk_Item *itemPtr, int prepass)
{
  register BLineItem *linePtr = (BLineItem *) itemPtr;
  char buffer[200];
  char *style;

  /*
   * Generate a path for the line's center-line (do this differently
   * for straight lines and smoothed lines).
   */

  Tk_CanvasPsPath(interp, canvas, linePtr->coordPtr, linePtr->numPoints);

  /*
   * Set other line-drawing parameters and stroke out the line.
   */

  sprintf(buffer, "%d setlinewidth\n", linePtr->width);
  Tcl_AppendResult(interp, buffer, (char *) NULL);
  style = "0 setlinecap\n";
  if (linePtr->capStyle == CapRound) 
  {
    style = "1 setlinecap\n";
  } 
  else if (linePtr->capStyle == CapProjecting) 
  {
    style = "2 setlinecap\n";
  }
  Tcl_AppendResult(interp, style, (char *) NULL);
  style = "0 setlinejoin\n";
  if (linePtr->joinStyle == JoinRound) 
  {
    style = "1 setlinejoin\n";
  } 
  else if (linePtr->joinStyle == JoinBevel) 
  {
    style = "2 setlinejoin\n";
  }
  Tcl_AppendResult(interp, style, (char *) NULL);
  
  if (Tk_CanvasPsColor(interp, canvas,
      Tk_3DBorderColor(linePtr->border)) != TCL_OK) 
  {
    return TCL_ERROR;
  }
  
  Tcl_AppendResult(interp, "stroke\n", (char *) NULL);

  return TCL_OK;
}

/*
 * Register the item...
 */
SCISHARE void BLineInit()
{
  Tk_CreateItemType(&TkBLineType);
}
