/*
   For more information, please see: http://software.sci.utah.edu

   The MIT License

   Copyright (c) 2009 Scientific Computing and Imaging Institute,
   University of Utah.

   
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included
   in all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
   OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
   THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
   DEALINGS IN THE SOFTWARE.
*/



/*
 *  RingWidget.cc
 *
 *  Written by:
 *   James Purciful
 *   Department of Computer Science
 *   University of Utah
 *   Jan. 1995
 *
 */

#include <Dataflow/Widgets/RingWidget.h>
#include <Dataflow/Constraints/AngleConstraint.h>
#include <Dataflow/Constraints/DistanceConstraint.h>
#include <Dataflow/Constraints/RatioConstraint.h>
#include <Core/Geom/GeomCylinder.h>
#include <Core/Geom/GeomSphere.h>
#include <Core/Geom/GeomGroup.h>
#include <Core/Geom/GeomTorus.h>
#include <Core/Geometry/Plane.h>

#include <Core/Math/MiscMath.h>
#include <Dataflow/Network/Module.h>

namespace SCIRun {



const Index NumCons = 6;
const Index NumVars = 8;
const Index NumGeoms = 10;
const Index NumPcks = 10;
const Index NumMatls = 5;
const Index NumSwtchs = 4;
const Index NumSchemes = 5;

enum { ConstRD, ConstRC, ConstDC, ConstPyth, ConstRadius, ConstAngle };
enum { GeomPointU, GeomPointR, GeomPointD, GeomPointL,
       GeomSlider, GeomRing,
       GeomResizeU, GeomResizeR, GeomResizeD, GeomResizeL };
enum { PickSphU, PickSphR, PickSphD, PickSphL,
       PickRing, PickSlider,
       PickResizeU, PickResizeR, PickResizeD, PickResizeL };

/***************************************************************************
 * The constructor initializes the widget's constraints, variables,
 *      geometry, picks, materials, modes, switches, and schemes.
 * Variables and constraints are initialized as a function of the
 *      widget_scale.
 * Much of the work is accomplished in the BaseWidget constructor which
 *      includes some consistency checking to ensure full initialization.
 */
RingWidget::RingWidget( Module* module, CrowdMonitor* lock, double widget_scale , bool is_slideable)
  : BaseWidget(module, lock, "RingWidget", NumVars, NumCons, NumGeoms, NumPcks, NumMatls, 0, NumSwtchs, widget_scale),
    oldrightaxis(1, 0, 0), olddownaxis(0, 1, 0),
    is_slideable_(is_slideable)
{
  const double INIT = 5.0*widget_scale;
  // Scheme4/5 are used to resize.
  variables[PointRVar] = new PointVariable("PntR", solve, Scheme1, Point(INIT, 0, 0));
  variables[PointDVar] = new PointVariable("PntD", solve, Scheme2, Point(0, INIT, 0));
  variables[CenterVar] = new PointVariable("Center", solve, Scheme1, Point(0, 0, 0));
  variables[SliderVar] = new PointVariable("Slider", solve, Scheme3, Point(INIT, 0, 0));
  variables[DistVar] = new RealVariable("Dist", solve, Scheme1, INIT);
  variables[HypoVar] = new RealVariable("HYPO", solve, Scheme1, sqrt(2.0*INIT*INIT));
  variables[Sqrt2Var] = new RealVariable("Sqrt2", solve, Scheme1, sqrt(2.0));
  variables[AngleVar] = new RealVariable("Angle", solve, Scheme3, 0);

  constraints[ConstAngle] = new AngleConstraint("ConstAngle",
						   NumSchemes,
						   variables[CenterVar],
						   variables[PointRVar],
						   variables[PointDVar],
						   variables[SliderVar],
						   variables[AngleVar]);
  constraints[ConstAngle]->VarChoices(Scheme1, 3, 3, 3, 3, 3);
  constraints[ConstAngle]->VarChoices(Scheme2, 3, 3, 3, 3, 3);
  constraints[ConstAngle]->VarChoices(Scheme3, 4, 4, 4, 4, 4);
  constraints[ConstAngle]->VarChoices(Scheme4, 3, 3, 3, 3, 3);
  constraints[ConstAngle]->VarChoices(Scheme5, 3, 3, 3, 3, 3);
  constraints[ConstAngle]->Priorities(P_Lowest, P_Lowest, P_Lowest,
				      P_Highest, P_Highest);
  constraints[ConstRadius] = new DistanceConstraint("ConstRadius",
						       NumSchemes,
						       variables[CenterVar],
						       variables[SliderVar],
						       variables[DistVar]);
  constraints[ConstRadius]->VarChoices(Scheme1, 1, 1, 1);
  constraints[ConstRadius]->VarChoices(Scheme2, 1, 1, 1);
  constraints[ConstRadius]->VarChoices(Scheme3, 1, 1, 1);
  constraints[ConstRadius]->VarChoices(Scheme4, 1, 1, 1);
  constraints[ConstRadius]->VarChoices(Scheme5, 1, 1, 1);
  constraints[ConstRadius]->Priorities(P_Lowest, P_HighMedium, P_Lowest);
  constraints[ConstRD] = new DistanceConstraint("ConstRD",
						   NumSchemes,
						   variables[PointRVar],
						   variables[PointDVar],
						   variables[HypoVar]);
  constraints[ConstRD]->VarChoices(Scheme1, 1, 1, 1);
  constraints[ConstRD]->VarChoices(Scheme2, 0, 0, 0);
  constraints[ConstRD]->VarChoices(Scheme3, 1, 0, 1);
  constraints[ConstRD]->VarChoices(Scheme4, 1, 1, 1);
  constraints[ConstRD]->VarChoices(Scheme5, 0, 0, 0);
  constraints[ConstRD]->Priorities(P_Default, P_Default, P_Default);
  constraints[ConstPyth] = new RatioConstraint("ConstPyth",
						  NumSchemes,
						  variables[HypoVar],
						  variables[DistVar],
						  variables[Sqrt2Var]);
  constraints[ConstPyth]->VarChoices(Scheme1, 0, 0, 0);
  constraints[ConstPyth]->VarChoices(Scheme2, 0, 0, 0);
  constraints[ConstPyth]->VarChoices(Scheme3, 0, 0, 0);
  constraints[ConstPyth]->VarChoices(Scheme4, 1, 0, 0);
  constraints[ConstPyth]->VarChoices(Scheme5, 1, 0, 0);
  constraints[ConstPyth]->Priorities(P_Highest, P_Highest, P_Highest);
  constraints[ConstRC] = new DistanceConstraint("ConstRC",
						   NumSchemes,
						   variables[PointRVar],
						   variables[CenterVar],
						   variables[DistVar]);
  constraints[ConstRC]->VarChoices(Scheme1, 0, 0, 0);
  constraints[ConstRC]->VarChoices(Scheme2, 0, 0, 0);
  constraints[ConstRC]->VarChoices(Scheme3, 0, 0, 0);
  constraints[ConstRC]->VarChoices(Scheme4, 2, 2, 2);
  constraints[ConstRC]->VarChoices(Scheme5, 0, 0, 0);
  constraints[ConstRC]->Priorities(P_Highest, P_Highest, P_Default);
  constraints[ConstDC] = new DistanceConstraint("ConstDC",
						   NumSchemes,
						   variables[PointDVar],
						   variables[CenterVar],
						   variables[DistVar]);
  constraints[ConstDC]->VarChoices(Scheme1, 0, 0, 0);
  constraints[ConstDC]->VarChoices(Scheme2, 0, 0, 0);
  constraints[ConstDC]->VarChoices(Scheme3, 0, 0, 0);
  constraints[ConstDC]->VarChoices(Scheme4, 0, 0, 0);
  constraints[ConstDC]->VarChoices(Scheme5, 2, 2, 2);
  constraints[ConstDC]->Priorities(P_Highest, P_Highest, P_Default);

  Index geom, pick;
  geometries[GeomRing] = new GeomTorus;
  picks_[PickRing] = new GeomPick(geometries[GeomRing], this, PickRing);
  picks(PickRing)->set_highlight(DefaultHighlightMaterial);
  materials[RingMatl] = new GeomMaterial(picks_[PickRing], DefaultEdgeMaterial);
  CreateModeSwitch(0, materials[RingMatl]);

  GeomGroup* pts = new GeomGroup;
  for (geom = GeomPointU, pick = PickSphU;
       geom <= GeomPointL; geom++, pick++)
  {
    geometries[geom] = new GeomSphere;
    picks_[pick] = new GeomPick(geometries[geom], this, pick);
    picks(pick)->set_highlight(DefaultHighlightMaterial);
    pts->add(picks_[pick]);
  }
  materials[PointMatl] = new GeomMaterial(pts, DefaultPointMaterial);
   
  geometries[GeomResizeL] = new GeomCappedCylinder;
  picks_[PickResizeL] = new GeomPick(geometries[GeomResizeL], this, PickResizeL);
  picks(PickResizeL)->set_highlight(DefaultHighlightMaterial);
  materials[HalfResizeMatl] = new GeomMaterial(picks_[PickResizeL], DefaultSpecialMaterial);

  GeomGroup* resulr = new GeomGroup;
  geometries[GeomResizeU] = new GeomCappedCylinder;
  picks_[PickResizeU] = new GeomPick(geometries[GeomResizeU],
				       this, PickResizeU);
  picks(PickResizeU)->set_highlight(DefaultHighlightMaterial);
  resulr->add(picks_[PickResizeU]);
  geometries[GeomResizeD] = new GeomCappedCylinder;
  picks_[PickResizeD] = new GeomPick(geometries[GeomResizeD],
				       this, PickResizeD);
  picks(PickResizeD)->set_highlight(DefaultHighlightMaterial);
  resulr->add(picks_[PickResizeD]);
  geometries[GeomResizeR] = new GeomCappedCylinder;
  picks_[PickResizeR] = new GeomPick(geometries[GeomResizeR],
				       this, PickResizeR);
  picks(PickResizeR)->set_highlight(DefaultHighlightMaterial);
  resulr->add(picks_[PickResizeR]);
  materials[ResizeMatl] = new GeomMaterial(resulr, DefaultResizeMaterial);

  GeomGroup* w = new GeomGroup;
  //w->add(materials[PointMatl]);
  w->add(materials[HalfResizeMatl]);
  w->add(materials[ResizeMatl]);
  CreateModeSwitch(1, materials[PointMatl]);
  CreateModeSwitch(2, w);

  geometries[GeomSlider] = new GeomCappedCylinder;
  picks_[PickSlider] = new GeomPick(geometries[GeomSlider], this, PickSlider);
  picks(PickSlider)->set_highlight(DefaultHighlightMaterial);
  materials[SliderMatl] = new GeomMaterial(picks_[PickSlider], DefaultSliderMaterial);
  CreateModeSwitch(3, materials[SliderMatl]);

  // Switch0 are the bars
  // Switch1 are the rotation points
  // Switch2 are the resize cylinders
  // Switch3 are the sliders
  if (is_slideable_)
  {
    SetNumModes(8);
    SetMode(Mode0, Switch0|Switch1|Switch2|Switch3);
    SetMode(Mode1, Switch0);
    SetMode(Mode2, Switch0|Switch1);
    SetMode(Mode3, Switch0|Switch2);
    SetMode(Mode4, Switch0|Switch1|Switch2);
    SetMode(Mode5, Switch0|Switch3);
    SetMode(Mode6, Switch0|Switch1|Switch3);
    SetMode(Mode7, Switch0|Switch2|Switch3);
  }
  else
  {
    SetNumModes(4);
    SetMode(Mode0, Switch0|Switch1|Switch2);
    SetMode(Mode1, Switch0);
    SetMode(Mode2, Switch0|Switch1);
    SetMode(Mode3, Switch0|Switch2);
  }

  FinishWidget();
}


/***************************************************************************
 * The destructor frees the widget's allocated structures.
 * The BaseWidget's destructor frees the widget's constraints, variables,
 *      geometry, picks, materials, modes, switches, and schemes.
 * Therefore, most widgets' destructors will not need to do anything.
 */
RingWidget::~RingWidget()
{
}


/***************************************************************************
 * The widget's redraw method changes widget geometry to reflect the
 *      widget's variable values and its widget_scale.
 * Geometry should only be changed if the mode_switch that displays
 *      that geometry is active.
 * Redraw should also set the principal directions for all picks.
 * Redraw should never be called directly; the BaseWidget execute method
 *      calls redraw after establishing the appropriate locks.
 */
void
RingWidget::redraw()
{
  Vector Right(GetRightAxis()*variables[DistVar]->real());
  Vector Down(GetDownAxis()*variables[DistVar]->real());
  Point Center(variables[CenterVar]->point());
  Point U(Center-Down);
  Point R(Center+Right);
  Point D(Center+Down);
  Point L(Center-Right);
   
  Vector normal(Cross(GetRightAxis(), GetDownAxis()));
  if (normal.length2() < 1e-6) { normal = Vector(1.0, 0.0, 0.0); }
   
  if (mode_switches[0]->get_state())
  {
    geometry<GeomTorus*>(GeomRing)->move(variables[CenterVar]->point(),
					 normal,
					 variables[DistVar]->real(),
					 0.5*widget_scale_);
  }

  if (mode_switches[1]->get_state())
  {
    geometry<GeomSphere*>(GeomPointU)->move(U, widget_scale_);
    geometry<GeomSphere*>(GeomPointR)->move(R, widget_scale_);
    geometry<GeomSphere*>(GeomPointD)->move(D, widget_scale_);
    geometry<GeomSphere*>(GeomPointL)->move(L, widget_scale_);
  }
  if (mode_switches[2]->get_state())
  {
    geometry<GeomCappedCylinder*>(GeomResizeU)->move(U, U - (GetDownAxis() * 1.5 * widget_scale_),
						     0.5*widget_scale_);
    geometry<GeomCappedCylinder*>(GeomResizeR)->move(R, R + (GetRightAxis() * 1.5 * widget_scale_),
						     0.5*widget_scale_);
    geometry<GeomCappedCylinder*>(GeomResizeD)->move(D, D + (GetDownAxis() * 1.5 * widget_scale_),
						     0.5*widget_scale_);
    geometry<GeomCappedCylinder*>(GeomResizeL)->move(L, L - (GetRightAxis() * 1.5 * widget_scale_),
						     0.5*widget_scale_);
  }

  Vector slide(variables[SliderVar]->point()-variables[CenterVar]->point());
  if (slide.length2() > 1e-6)
  {
    slide = Cross(normal, slide.safe_normal());
    if (slide.length2() < 1e-6)  { slide = Vector(0.0, 1.0, 0.0); }
  }
  else
  {
    slide = GetRightAxis();
  }
  if (mode_switches[3]->get_state())
  {
    geometry<GeomCappedCylinder*>(GeomSlider)->
      move(variables[SliderVar]->point() - (slide * 0.3 * widget_scale_),
	   variables[SliderVar]->point() + (slide * 0.3 * widget_scale_),
	   1.1*widget_scale_);
  }
   
  ((DistanceConstraint*)constraints[ConstRC])->SetMinimum(1.6*widget_scale_);
  ((DistanceConstraint*)constraints[ConstDC])->SetMinimum(1.6*widget_scale_);
  ((DistanceConstraint*)constraints[ConstRD])->SetMinimum(sqrt(2*1.6*1.6)*widget_scale_);

  Right.safe_normalize();
  Down.safe_normalize();
  Vector Norm(Cross(Right, Down));
  for (Index geom = 0; geom < NumPcks; geom++)
  {
    if (geom == PickSlider)
    {
      picks(geom)->set_principal(slide);
    }
    else if ((geom == PickResizeU) || (geom == PickResizeD))
    {
      picks(geom)->set_principal(Down);
    }
    else if ((geom == PickResizeL) || (geom == PickResizeR))
    {
      picks(geom)->set_principal(Right);
    }
    else if ((geom == PickSphR) || (geom == PickSphL))
    {
      picks(geom)->set_principal(Down, Norm);
    }
    else if ((geom == PickSphU) || (geom == PickSphD))
    {
      picks(geom)->set_principal(Right, Norm);
    }
    else
    {
      picks(geom)->set_principal(Right, Down, Norm);
    }
  }
}


void
RingWidget::geom_pick( GeomPickHandle p,
		       ViewWindow*vw, int data, const BState& bs)
{
  BaseWidget::geom_pick(p, vw, data, bs);
  pick_pointrvar_ = variables[PointRVar]->point();
  pick_pointdvar_ = variables[PointDVar]->point();
  pick_centervar_ = variables[CenterVar]->point();
  pick_slidervar_ = variables[SliderVar]->point();
}


/***************************************************************************
 * The widget's geom_moved method receives geometry move requests from
 *      the widget's picks.  The widget's variables must be altered to
 *      reflect these changes based upon which pick made the request.
 * No more than one variable should be Set since this triggers solution of
 *      the constraints--multiple Sets could lead to inconsistencies.
 *      The constraint system only requires that a variable be Set if the
 *      change would cause a constraint to be invalid.  For example, if
 *      all PointVariables are moved by the same delta, then no Set is
 *      required.
 * The last line of the widget's geom_moved method should call the
 *      BaseWidget execute method (which calls the redraw method).
 */
void
RingWidget::geom_moved( GeomPickHandle, int axis, double dist,
			const Vector& delta, int pick, const BState&,
			const Vector &pick_offset)
{
  Point p;
  const double resize_min(1.5*widget_scale_);
  if (axis==1) dist = -dist;

  ((DistanceConstraint*)constraints[ConstRC])->SetDefault(GetRightAxis());
  ((DistanceConstraint*)constraints[ConstDC])->SetDefault(GetDownAxis());

  switch(pick)
  {
  case PickSphU:
    variables[PointRVar]->Move(pick_pointrvar_);
    variables[PointDVar]->Move(pick_pointdvar_);
    variables[CenterVar]->Move(pick_centervar_);
    variables[SliderVar]->Move(pick_slidervar_);
    variables[PointDVar]->SetDelta(-pick_offset);
    break;

  case PickSphR:
    variables[PointRVar]->Move(pick_pointrvar_);
    variables[PointDVar]->Move(pick_pointdvar_);
    variables[CenterVar]->Move(pick_centervar_);
    variables[SliderVar]->Move(pick_slidervar_);
    variables[PointRVar]->SetDelta(pick_offset);
    break;

  case PickSphD:
    variables[PointRVar]->Move(pick_pointrvar_);
    variables[PointDVar]->Move(pick_pointdvar_);
    variables[CenterVar]->Move(pick_centervar_);
    variables[SliderVar]->Move(pick_slidervar_);
    variables[PointDVar]->SetDelta(pick_offset);
    break;

  case PickSphL:
    variables[PointRVar]->Move(pick_pointrvar_);
    variables[PointDVar]->Move(pick_pointdvar_);
    variables[CenterVar]->Move(pick_centervar_);
    variables[SliderVar]->Move(pick_slidervar_);
    variables[PointRVar]->SetDelta(-pick_offset);
    break;

  case PickResizeU:
    if ((variables[DistVar]->real() - dist) < resize_min)
    {
      p = variables[CenterVar]->point() + GetDownAxis()*resize_min;
    }
    else
    {
      p = variables[PointDVar]->point() - delta;
    }
    variables[PointDVar]->Set(p, Scheme5);
    break;

  case PickResizeR:
    if ((variables[DistVar]->real() + dist) < resize_min)
    {
      p = variables[CenterVar]->point() + GetRightAxis()*resize_min;
    }
    else
    {
      p = variables[PointRVar]->point() + delta;
    }
    variables[PointRVar]->Set(p, Scheme4);
    break;

  case PickResizeD:
    if ((variables[DistVar]->real() + dist) < resize_min)
    {
      p = variables[CenterVar]->point() + GetDownAxis()*resize_min;
    }
    else
    {
      p = variables[PointDVar]->point() + delta;
    }
    variables[PointDVar]->Set(p, Scheme5);
    break;

  case PickResizeL:
    if ((variables[DistVar]->real() - dist) < resize_min)
    {
      p = variables[CenterVar]->point() + GetRightAxis()*resize_min;
    }
    else
    {
      p = variables[PointRVar]->point() - delta;
    }
    variables[PointRVar]->Set(p, Scheme4);
    break;

  case PickSlider:
    if (is_slideable_)
    {
      variables[SliderVar]->SetDelta(delta);
    }
    break;

  case PickRing:
    variables[PointRVar]->Move(pick_pointrvar_);
    variables[PointDVar]->Move(pick_pointdvar_);
    variables[CenterVar]->Move(pick_centervar_);
    variables[SliderVar]->Move(pick_slidervar_);
    MoveDelta(pick_offset);
    break;
  }
  execute(0);
}


/***************************************************************************
 * This standard method simply moves all the widget's PointVariables by
 *      the same delta.
 * The last line of this method should call the BaseWidget execute method
 *      (which calls the redraw method).
 */
void
RingWidget::MoveDelta( const Vector& delta )
{
  variables[PointRVar]->MoveDelta(delta);
  variables[PointDVar]->MoveDelta(delta);
  variables[CenterVar]->MoveDelta(delta);
  variables[SliderVar]->MoveDelta(delta);

  execute(1);
}


/***************************************************************************
 * This standard method returns a reference point for the widget.  This
 *      point should have some logical meaning such as the center of the
 *      widget's geometry.
 */
Point
RingWidget::ReferencePoint() const
{
  return variables[CenterVar]->point();
}


void
RingWidget::SetPosition( const Point& center, const Vector& normal, const double radius )
{
  Vector v1, v2;
  normal.find_orthogonal(v1, v2);
  variables[CenterVar]->Move(center);
  variables[PointRVar]->Move(center+v1*radius);
  variables[PointDVar]->Set(center+v2*radius);
  SetRadius(radius);

  execute(0);
}


void
RingWidget::GetPosition( Point& center, Vector& normal, double& radius ) const
{
  center = variables[CenterVar]->point();
  normal = Plane(variables[PointRVar]->point(),
		 variables[PointDVar]->point(),
		 variables[CenterVar]->point()).normal();
  radius = variables[DistVar]->real();
}


void
RingWidget::SetRatio( const double ratio )
{
  ASSERT((ratio>=0.0) && (ratio<=1.0));
  variables[AngleVar]->Set(ratio*2.0*M_PI - M_PI);

  execute(0);
}


double
RingWidget::GetRatio() const
{
  double ratio=variables[AngleVar]->real() / (2.0 * M_PI);
  if (ratio < 0)
  {
    ratio+=M_PI;
  }
  return ratio;
}


void
RingWidget::SetRadius( const double radius )
{
  ASSERT(radius>=0.0);
   
  Vector axis1(variables[PointRVar]->point() - variables[CenterVar]->point());
  Vector axis2(variables[PointDVar]->point() - variables[CenterVar]->point());
  double ratio(radius/variables[DistVar]->real());

  variables[PointRVar]->Move(variables[CenterVar]->point()+axis1*ratio);
  variables[PointDVar]->Move(variables[CenterVar]->point()+axis2*ratio);

  variables[DistVar]->Set(variables[DistVar]->real()*ratio); // This should set the slider...

  execute(0);
}


double
RingWidget::GetRadius() const
{
  return variables[DistVar]->real();
}


const Vector&
RingWidget::GetRightAxis()
{
  Vector axis(variables[PointRVar]->point() - variables[CenterVar]->point());
  if (axis.length2() <= 1e-6)
  {
    return oldrightaxis;
  }
  else
  {
    return (oldrightaxis = axis.safe_normal());
  }
}


const Vector&
RingWidget::GetDownAxis()
{
  Vector axis(variables[PointDVar]->point() - variables[CenterVar]->point());
  if (axis.length2() <= 1e-6)
  {
    return olddownaxis;
  }
  else
  {
    return (olddownaxis = axis.safe_normal());
  }
}


void
RingWidget::GetPlane(Vector& v1, Vector& v2)
{
  v1=GetRightAxis();
  v2=GetDownAxis();
}


/***************************************************************************
 * This standard method returns a string describing the functionality of
 *      a widget's material property.  The string is used in the 
 *      BaseWidget UI.
 */
std::string
RingWidget::GetMaterialName( const Index mindex ) const
{
  ASSERT(mindex<materials.size());
   
  switch(mindex)
  {
  case 0:
    return "Point";
  case 1:
    return "Ring";
  case 2:
    return "Slider";
  case 3:
    return "Resize";
  case 4:
    return "HalfResize";
  default:
    return "UnknownMaterial";
  }
}


std::string
RingWidget::GetStateString()
{
  std::string base = BaseWidget::GetStateString();

  Point center;
  Vector normal;
  double radius;
  GetPosition(center, normal, radius);

  std::ostringstream os;
  os << " " << center.x() << " " << center.y() << " " << center.z();
  os << " " << normal.x() << " " << normal.y() << " " << normal.z();
  os << " " << radius;

  return base + os.str();
}


int
RingWidget::SetStateString(const std::string &str)
{
  const int offset = BaseWidget::SetStateString(str);
  std::istringstream is(str.substr(offset));

  double centerx, centery, centerz;
  double normalx, normaly, normalz;
  double radius;
  is >> centerx >> centery >> centerz;
  is >> normalx >> normaly >> normalz;
  is >> radius;

  if (!is.fail())
  {
    SetPosition(Point(centerx, centery, centerz),
		Vector(normalx, normaly, normalz),
		radius);
  }

  return offset + is.tellg();
}


} // End namespace SCIRun


